#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
########################################################################
#
# Plot the subprocess decomposition
#
# Created by B. Schillinger, 01.06.2018
# Modified by K. Rabbertz, 15.04.2019
# Simplified, standardised, and prepared for python3 by K. Rabbertz, 06/07 2024
#
########################################################################
#
# fastNLO for direct evaluation of interpolation grids
# ATTENTION: fastNLO python extension is required for Python 3!
from fastnlo import SetGlobalVerbosity
from fastnlo import fastNLOLHAPDF
import fastnlo
import matplotlib.pyplot as plt
import argparse
import os
import re
import numpy as np
import timeit
import matplotlib as mpl
import matplotlib.gridspec as gridspec
from matplotlib.ticker import (FormatStrFormatter, LogFormatter, NullFormatter, ScalarFormatter, AutoMinorLocator, MultipleLocator)
from matplotlib.pyplot import cm
# import warnings
# warnings.filterwarnings("error")


# We do not want any interactive plotting! Figures are saved to files instead.
# This also avoids the ANNOYANCE of frequently missing Tkinter/tkinter (python2/3) GUI backends!
# To produce scalable graphics for publication use eps, pdf, or svg as file format.
# For this to work we try the Cairo backend, which can do all of these plus the raster format png.
# If this is not usable, we fall back to the Agg backend capable only of png for nice web plots.
#ngbackends = mpl.rcsetup.non_interactive_bk
#print('[fastnnlo_scaleunc]: Non GUI backends are: ', ngbackends)
# 1st try cairo
backend = 'cairo'
usecairo = True
try:
    import cairocffi as cairo
except ImportError:
    try:
        import cairo
    except ImportError:
        usecairo = False
    else:
        if cairo.version_info < (1, 11, 0):
            # Introduced create_for_data for Py3.
            usecairo = False
if usecairo:
    mpl.use('cairo')
else:
    backend = 'agg'
    useagg = True
    try:
        mpl.use(backend, force=True)
        print('[fnlo-py-subprocs]: Warning! Could not import cairo backend :-( Using agg instead for raster plots only!')
    except:
        useagg = False
        print('[fnlo-py-subprocs]: Can not use agg backend :-(')
        raise ImportError('[fnlo-py-subprocs]: Neither cairo nor agg backend found :-( Cannot produce any plots. Good bye!')
    mpl.use('agg')


# Redefine ScalarFormatter
class ScalarFormatterForceFormat(ScalarFormatter):
    # Override function that finds format to use.
    def _set_format(self, vmin, vmax):
        self.format = "%1.2f"  # Give format here


# Action class to allow comma-separated (or empty) list in options
class SplitArgs(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if values:
            setattr(namespace, self.dest, values[0].split(','))
        else:
            setattr(namespace, self.dest, [''])


# Some global definitions
_fntrans = str.maketrans({'[': '', ']': '', '(': '', ')': '', ',': ''})  # Filename translation table
_formats = {'eps': 0, 'pdf': 1, 'png': 2, 'svg': 3}
_text_to_order = {'LO': 0, 'NLO': 1, 'NNLO': 2}
_order_to_text = {0: 'LO', 1: 'NLO', 2: 'NNLO'}

# Some process dictionaries
_process_color = {"gg": "red", "gq": "orange", "gaq": "yellow",
                  "qq": "green", "qaq": "blue", "aqaq": "purple"}
_process_names = {
    "gg":   "Gluon-Gluon",
    "gq":   "Gluon-Quark",
    "gaq":  "Gluon-Antiquark",
    "qq":   "Quark-Quark",
    "qaq":  "Quark-Antiquark",
    "aqaq": "Antiquark-Antiquark",
}
_process_texnames = {
    "gg": r"$\mathrm{gg}$",
    "gq": r"$\mathrm{gq}$",
    "gaq": r"$\mathrm{g\overline{q}}$",
    "qq": r"$\mathrm{qq}$",
    "qaq": r"$\mathrm{q\overline{q}}$",
    "qaq": r"$\mathrm{q\overline{q}}$",
    "aqaq": r"$\mathrm{\overline{q}\overline{q}}$",
}

#####################################################################################


def main():

    # Start timer
    # just for measuring wall clock time - not necessary
    start_time = timeit.default_timer()

    # Define arguments & options
    parser = argparse.ArgumentParser(epilog='', formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    # Positional arguments
    parser.add_argument('tabfile', type=argparse.FileType('r'),
                        help='fastNLO table to be evaluated including path and extension .tab or tab.gz.')

    # Optional arguments
    parser.add_argument('-s', '--subprocs', required=False, nargs=1, type=str, action=SplitArgs,
                        default=['gg', 'gq', 'gaq', 'qq', 'qaq', 'aqaq'],
                        choices=['gg', 'gq', 'gaq', 'qq', 'qaq', 'aqaq'],
                        help='Subprocesses to be compared.\n'
                        'Options: gg, gq, gaq, qq, qaq, aqaq (TODO: q=q, q!q, uu, uau, auau, etc.)\n'
                        'See SelectProcesses() in fastNLOReader.cc for more information.')

    parser.add_argument('-o', '--order', required=False, nargs='?', type=str, default='NNLO', choices=['LO', 'NLO', 'NNLO'],
                        help='Order to which the subprocess contributions are calculated.')

    parser.add_argument('-n', '--norm-order', required=False, nargs='?', type=str, default='SAME', choices=['LO', 'NLO', 'NNLO', 'SAME'],
                        help='Normalisation order to which the subprocesses contributions are normalised.')

    parser.add_argument('-p', '--pdfset', default='CT18NNLO',
                        help='PDFset to use with fastNLO table.')

    parser.add_argument('-m', '--member', default=0, type=int,
                        help='Member of PDFset to use, default is 0.')

    parser.add_argument('-v', '--verbosity', type=int, default=1, choices=[-1000, 0, 1, 2, 1000],
                        help='Set verbosity of fastNLO. \n'
                        'Choices are -1000 (DEBUG), 0 (INFO), 1 (WARNING), 2 (ERROR), 1000 (SILENT)')
    # Options for plotting details
    parser.add_argument('--filename', default=None, type=str,
                        help='Replace default filename derived from table name and orders by given string.')
    parser.add_argument('--format', required=False, nargs=1, type=str, action=SplitArgs,
                        help='Comma-separated list of plot formats to use: eps, pdf, png, svg. If nothing is chosen, png is used.')
    parser.add_argument('--title', default=None, type=str,
                        help='Replace table name as default title by given string.')
    parser.add_argument('--legtitle', default=None, type=str,
                        help='Replace order & PDF set as default title by given string.')
    parser.add_argument('--xlabel', default=None, type=str,
                        help='Replace default x axis label from table by given string, e.g. \'$p_\mathrm{T}$ [GeV]\'')
    parser.add_argument('--ylabel', default='Subprocess decomposition', type=str,
                        help='Replace y axis label by given string.')
    parser.add_argument('--xmin', default=None, type=float,
                        help='Replace default x axis minimum from table by given number.')
    parser.add_argument('--xmax', default=None, type=float,
                        help='Replace default x axis maximum from table by given number.')
    parser.add_argument('--ymin', default=None, type=float,
                        help='Replace default y axis minimum from table by given number.')
    parser.add_argument('--ymax', default=None, type=float,
                        help='Replace default y axis maximum from table by given number.')

    # Print header
    print("\n###########################################################################################")
    print("# fnlo_py_subprocs:")
    print("# Plot the subprocess decomposition")
    print("###########################################################################################\n")

    # Parse arguments
    args = vars(parser.parse_args())

    # Table and table name
    table = args['tabfile'].name
    tabbase = os.path.basename(table)
    # Eliminate extensions tab and gz
    tabname = re.sub('.gz$', '', tabbase)
    tabname = re.sub('.tab$', '', tabname)
    print('[fnlo-py-subprocs]: Going to read table {}.'.format(tabbase))

    # PDF set
    pdfset_ = os.path.basename(args['pdfset'])
    pdfname = os.path.splitext(pdfset_)[0]
    print('[fnlo-py-subprocs]: Using PDF set {}.'.format(pdfname))

    # Subprocesses to investigate
    print('[fnlo-py-subprocs]: Subprocesses that will be investigated: ')
    print('[fnlo-py-subprocs]: ', args['subprocs'])

    # Set subprocess and normalisation orders
    order = args['order']
    norm_order = args['norm_order']
    if norm_order == 'SAME':
        norm_order = order
    print('[fnlo-py-subprocs]: The chosen subprocess order is: {}'.format(order))
    print('[fnlo-py-subprocs]: The chosen normalisation order is: {}'.format(norm_order))

    # Plot formats to use
    formats = args['format']
    if formats is None:
        formats = ['png']
    for fmt in formats:
        if fmt not in _formats.keys():
            print('[fnlo-py-subprocs]: Illegal format specified, aborted!')
            print('[fnlo-py-subprocs]: Format list:', args['format'])
            exit(1)

    # Verbosity
    ldebug = False
    if args['verbosity'] < 1:
        ldebug = True

    ###################### Start EVALUATION with fastNLO library ###################################################
    print('[fnlo-py-subprocs]: Starting table evaluation.')
    SetGlobalVerbosity(args['verbosity'])
    fnlo = fastnlo.fastNLOLHAPDF(table, args['pdfset'], args['member'])

    # Get info from table
    # Get dimensionality of table:
    ndim = fnlo.GetNumDiffBin()
    if ndim > 1:
        print('[fnlo-py-subprocs]: Error! Script only tested for use with 1-dim. grids, but ndim = {}. Aborted!'.format(ndim))

    # Determine maximum of subsequent orders available in table
    max_iorder = -1
    o_existence = []
    for i in range(_text_to_order['NNLO']+1):
        o_existence.append(fnlo.SetContributionON(fastnlo.kFixedOrder, i, True))
        if o_existence[i]:
            max_iorder = max_iorder + 1
        else:
            break
    print('[fnlo-py-subprocs]: Maximum subsequent order found in table is: {}, corresponding to {}.'.format(max_iorder, _order_to_text[max_iorder]))
    if max_iorder < _text_to_order[order] or max_iorder < _text_to_order[norm_order]:
        print('[fnlo-py-subprocs]: Error! Maximum subsequent order available smaller than requested order. Aborted!')
        print('[fnlo-py-subprocs]: Requested: {}, found in table: {} vs. {}'.format(max_iorder, _text_to_order[order], _text_to_order[norm_order]))
        exit(1)

    # Get labels of all dimensions in table
    labels = fnlo.GetDimLabels()
    # Get x label from first dimension:
    xlabel = fnlo.GetDimLabel(0)
    if args['xlabel'] is not None:
        xlabel = args['xlabel']
    print("[fnlo-py-subprocs]: x-axis label: {}".format(xlabel))
    # y label
    ylabel = args['ylabel']
    print("[fnlo-py-subprocs]: y-axis label: {}".format(ylabel))
    # Plot title
    title = tabname
    if args['title'] is not None:
        title = args['title']
    print("[fnlo-py-subprocs]: Plot title: {}".format(title))

    # Get bin bounds in x
    bin_bounds = np.array(fnlo.GetObsBinsBounds(0))
    xlow = bin_bounds[0][0]
    xhig = bin_bounds[-1][1]

    # Set x axis limits
    if args['xmin'] is not None:
        xlow = args['xmin']
    if args['xmax'] is not None:
        xhig = args['xmax']
    print("[fnlo-py-subprocs]: x-axis minimum: {}".format(xlow))
    print("[fnlo-py-subprocs]: x-axis maximum: {}".format(xhig))

    # Set y axis limits (ylow, yhig might be None, i.e. unset!)
    ylow = args['ymin']
    yhig = args['ymax']

    # Evaluate cross section for normalisation
    print("[fnlo-py-subprocs]: Normalisation order: {}".format(norm_order))
    for j in range(0, max_iorder+1):
        fnlo.SetContributionON(fastnlo.kFixedOrder, j, j <= _text_to_order[norm_order])
    fnlo.CalcCrossSection()
    xs_all_oneplot = np.array(fnlo.GetCrossSection())
    if ldebug:
        print("[fnlo-py-subprocs]: The normalisation cross section is:")
        print("[fnlo-py-subprocs]: ", xs_all_oneplot)

    print("[fnlo-py-subprocs]: Subprocess Order: {}".format(order))
    for l in range(0, max_iorder+1):
        fnlo.SetContributionON(fastnlo.kFixedOrder, l, l <= _text_to_order[order])
    xs_subproc_list = []
    for i in range(0, len(args['subprocs'])):
        subproc_name = args['subprocs'][i]
        print('[fnlo-py-subprocs]: Selected subprocess: {}'.format(subproc_name))
        fnlo.SelectProcesses(subproc_name, True)
        fnlo.CalcCrossSection()
        xs_subproc = np.array(fnlo.GetCrossSection())
        xs_subproc_list.append(xs_subproc)
    xs_sub_oneplot = np.array(xs_subproc_list)

    # Calculate fraction of subprocess at order on total cross sectio at norm_order
    fractions_sub_oneplot = []
    for k in range(0, len(args['subprocs'])):
        fractions_sub_oneplot.append(np.divide(xs_sub_oneplot[k, :], xs_all_oneplot[:]))
    fractions_oneplot = np.array(fractions_sub_oneplot)

    # Checking whether contribution is positive or negative
    pos_contr = np.array(fractions_oneplot)
    neg_contr = np.array(fractions_oneplot)

    # Go through 'fractions' = subprocess after subprocess
    for i in range(0, len(args['subprocs'])):
        for j in range(0, len(bin_bounds)):  # Check the different bins
            if (fractions_oneplot[i, j] < 0):
                pos_contr[i, j] = 0.0
            else:
                neg_contr[i, j] = 0.0
        if ldebug:
            print('[fnlo-py-subprocs]: Positive contributions of {} in {} on total xs in {}:'.format(args['subprocs'][i], order, norm_order))
            print('[fnlo-py-subprocs]: ', pos_contr[i, :], '\n')
            print('[fnlo-py-subprocs]: Negative contributions of {} in {} on total xs in {}:'.format(args['subprocs'][i], order, norm_order))
            print('[fnlo-py-subprocs]: ', neg_contr[i, :], '\n')

    if ldebug:
        print('[fnlo-py-subprocs]: Summary of positive and negative contributions of the subprocesses:\n')
        print('[fnlo-py-subprocs]: Positive:', pos_contr, '\n')
        print('[fnlo-py-subprocs]: Negative:', neg_contr, '\n')

    # Create stackplot that compares the contribution of the subprocesses to the total cross section
    plt.close()

    fig = plt.figure(figsize=(8, 7))
    ax0 = fig.add_subplot(111)
    # Number of bins via bin_bounds variable
    # For tracking the current 'height' of stackplot, in order to staple and not overlap the values for each subprocess
    y0_pos = np.zeros([len(bin_bounds)])
    # Same as above for the negative contributions, plotted below x-axis
    y0_neg = np.zeros([len(bin_bounds)])
    patches = []  # Needed later for legend

    # Go through fractions, subproc by subproc
    for i in range(0, len(args['subprocs'])):
        cl = _process_color[args['subprocs'][i]]
        lb = _process_texnames[args['subprocs'][i]]

        # Positive contributions
        ax0.fill_between(bin_bounds.flatten(), steppify_bin(y0_pos), steppify_bin(y0_pos+pos_contr[i, :]), facecolor=cl, alpha=0.75)
        # Calculate new height of stackplot
        y0_pos += pos_contr[i, :]

        # Negative contributions
        ax0.fill_between(bin_bounds.flatten(), steppify_bin(y0_neg), steppify_bin(y0_neg+neg_contr[i, :]), facecolor=cl, hatch='X', alpha=0.75)
        # Calculate new height below x-axis
        y0_neg += neg_contr[i, :]

        # Patch for subprocess i (for the legend)
        patches.append(mpl.patches.Rectangle((0, 0), 0, 0, color=cl, label=lb, alpha=0.75))

    # Derive y-axis limits from content, if not set otherwise
    if ylow is None:
        ylow = np.amin(y0_neg)
    if yhig is None:
        yhig = np.amax(y0_pos)
    print("[fnlo-py-subprocs]: y-axis minimum: {}".format(ylow))
    print("[fnlo-py-subprocs]: y-axis maximum: {}".format(yhig))

    # Set axes limits
    ax0.set_xlim(xlow, xhig)
    ax0.set_ylim(ylow, yhig)
    ax0.set_xscale('log', nonpositive='clip')
    # ax0.autoscale(enable=True, axis='x', tight=True)

    for i in range(len(args['subprocs'])-1, -1, -1):
        ax0.add_patch(patches[i])

    # Dotted line at xs_sub/xs = 1 = 100%
    plt.axhline(y=1, xmin=0, xmax=1, color='k', linestyle='dotted')
    # Dotted line at xs_sub/xs = 0 = 0%
    plt.axhline(y=0, xmin=0, xmax=1, color='k', linestyle='-')

    plt.title(label=title, fontsize=16, loc='center')
    ax0.set_xlabel(xlabel, fontsize=16, loc='right')
    ax0.set_ylabel(ylabel, fontsize=16, loc='top')
    plt.xticks(fontsize=16)
    plt.yticks(fontsize=16)
    # TODO: ticks & labels
    # ax0.get_xaxis().set_major_locator(mpl.ticker.LogLocator(subs=(1.0, 2.0, 4.0, 7.0)))
    # ax0.get_xaxis().set_major_formatter(mpl.ticker.LogFormatter())
    legtitle = '{} / {}'.format(order, norm_order)
    if order == norm_order:
        legtitle = '{}'.format(order)
    legtitle = legtitle+'\n'+args['pdfset']
    if args['legtitle'] is not None:
        legtitle = args['legtitle']
    plt.legend(title=legtitle, loc='best')  # reverse=True requires matplotlib 3.7

    plt.tight_layout()
    fig.tight_layout()

    # Naming of the plots
    figname = tabname
    if args['filename'] is not None:
        figname = args['filename']
    figname = figname + '_' + order + '-vs-' + norm_order

    # Store in all requested formats
    for fmt in formats:
        filename = '%s.%s' % (figname, fmt)
        # ban characters defined in _fntrans for filenames (parentheses and commas)
        filename = filename.translate(_fntrans)
        fig.savefig(filename, bbox_inches='tight')
        print('[fnlo-py-subprocs]: Plot saved as:', filename)
    plt.close(fig)

    stop_time = timeit.default_timer()
    timediff = stop_time-start_time
    print('fnlo-py-subprocs: Elapsed time: %s sec = %s min' % (timediff, round(timediff/60, 2)))


##################################################
# function to make better uncertainty plot (shaded)
# could maybe be replaced by .flatten() ??


def steppify_bin(arr, isx=False):
    """
    Produce stepped array of arr, needed for example for stepped fill_betweens.
    Pass all x bin edges to produce stepped x arr and all y bincontents to produce
    stepped bincontents representation
    steppify_bin([1,2,3], True)
    -> [1,2,2,3]
    steppify_bin([5,6])
    -> [5,5,6,6]
    """
    if isx:
        # newarr = np.array(zip(arr[:-1], arr[1:])).ravel()
        newarr = np.array(list(zip(arr[:-1], arr[1:]))).flatten()
    else:
        # newarr = np.array(zip(arr, arr)).ravel()
        newarr = np.array(list(zip(arr, arr))).flatten()
    return newarr


#################################################

if __name__ == '__main__':
    main()
